/***********************************************************
        Copyright 1996 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/



/*
 * This module reads and interprets the configuration information for dynamic
 * address assignment.  (dhcpd.conf)
 */

#include <stdio.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/stat.h>

#include "hash.h"
#include "bootp.h"
#include "bootpd.h"
#include "dhcp.h"
#include "dhcpd.h"
#include "readfile.h"
#include "perm.h"
#include "dynamic.h"

#define KEYW_NULL		   0
#define KEYW_GATEWAY		   1
#define KEYW_DUMP		   2
#define KEYW_STATIC		   3
#define KEYW_PRIMARY		   4
#define KEYW_SECONDARY		   5
#define KEYW_NETWORK		   6
#define KEYW_DIRECTORY		   7
#define KEYW_FLUSHITEMS		   8
#define KEYW_FLUSHINTERVAL		   9
#define KEYW_BINDINGS		   10
#define KEYW_GCINTERVAL		   11

#define MAXTOKENS		 256	/* Max tokens per line */
#define MAXENTRYLEN		1024	/* Max line length */




extern void     report();
extern void     eat_whitespace();
extern void     makelower();
extern int      prs_inetaddr();

struct keywordmap {
	char           *keyword;
	int             symbolcode;
};


/*
 * List of keywords used in the dhcpd.conf file.  The order and actual values
 * of the symbol codes (KEYW_. . .) are unimportant, but they must all be
 * unique.  Keywords must be listed in all lowercase.
 */

PRIVATE struct keywordmap keyword_list[] = {
#ifdef DEBUG
	{"dump", KEYW_DUMP},
#endif
	{"gateway", KEYW_GATEWAY},
	{"network", KEYW_NETWORK},
	{"gcinterval", KEYW_GCINTERVAL},
	{"flushitems", KEYW_FLUSHITEMS},
	{"flushinterval", KEYW_FLUSHINTERVAL},
	{"binddir", KEYW_BINDINGS},
};





/*
 * Forward declarations
 */

PRIVATE void    kw_gateway();
PRIVATE void    kw_network();
PRIVATE int 	parse_line();
PRIVATE char          **get_tokens();
char           *savestr();


/*
 * Global variables
 */





#ifndef CONFIG_FILE
#define CONFIG_FILE "/etc/dhcpd.conf"
#endif
hash_tbl * cablehshtbl;

hash_tbl       *gwhshtbl;
char           *conffile = CONFIG_FILE;
unsigned long   conftime = 0;
void 
readconf(force)
	int             force;
{
	FILE           *fp;
	struct stat     st;
	if (init) {
		cablehshtbl = hash_Init(HASHTABLESIZE);
		gwhshtbl = hash_Init(HASHTABLESIZE);
		hash_Reset(cablehshtbl, free_cable);
		hash_Reset(gwhshtbl, free_gw);
	}
	/*
	 * Check the last modification time.
	 */
	if (stat(conffile, &st) < 0) {
		if(debug)
		report(LOG_ERR, "stat on \"%s\": %s", conffile, get_errmsg());
		return;
	}
	if ((force == 0) && (st.st_mtime == conftime) && st.st_nlink) {
		/*
		 * hasn't been modified or deleted yet.
		 */
		return;
	}
	conftime = st.st_mtime;

	if (debug)
		report(LOG_INFO, "reading %s\"%s\"",
		       (conftime != 0L) ? "new " : "", conffile);


	if ((fp = fopen(conffile, "r")) == NULL) {
		report(LOG_ERR, "unable to read configuration file \"%s\"\n",
		       conffile);
		exit(1);
	}
	hash_Reset(cablehshtbl, free_cable);
	hash_Reset(gwhshtbl, free_gw);
	while (parse_line(fp) >= 0);
	fclose(fp);
	if (debug > 1)
		report(LOG_INFO, "Finished reading %s", conffile);
}

void 
free_cable(cable)
	struct cable   *cable;
{
	struct link_list *t;
	if (!cable)
		return;
	if (--cable->linkcount)
		return;

	del_string(cable->name);

	while (cable->head) {
		t = cable->head;
		cable->head = cable->head->next;
		free(t);
	}
	free(cable);

}

void 
del_string_list(str)
	struct shared_string_list *str;
{
	if (!str)
		return;
	if (--str->linkcount)
		return;
	while (str->num)
		free(str->str[str->num--]);
	free(str);
}

void 
free_gw(gw)
	struct gateway *gw;
{
	int             i;
	if (!gw)
		return;
	if (--gw->linkcount)
		return;
	for (i = 0; i < gw->num; i++)
		free(gw->nets[i]);
	free(gw->nets);
	free(gw);
}





/*
 * This function reads a line of text from the given file and interprets it.
 * The first token (word) on the line is taken as a keyword which determines
 * the appropriate action.
 * 
 * Successful calls return 0.  A value of -1 indicates end-of-file.
 */

PRIVATE int 
parse_line(fp)
	FILE           *fp;
{
	static char     linebuf[MAXENTRYLEN], tmpstring[MAXENTRYLEN];
	struct keywordmap *keywd;
	char          **keyvec;
	int             i;

	if (fgets(linebuf, sizeof(linebuf), fp) == NULL) {
		return -1;
	}
	keyvec = get_tokens(linebuf);
	if (*keyvec) {
		if ((**keyvec == '#') || (**keyvec == ';')) {
			/*
			 * Comment, just return.
			 */
			return 0;
		}
		strcpy(tmpstring, *keyvec);
		makelower(tmpstring);
		keywd = keyword_list;
		for (i = sizeof(keyword_list) / sizeof(struct keywordmap); i > 0; i--) {
			if (!strcmp(keywd->keyword, tmpstring)) {
				break;
			}
			keywd++;
		}
		if (i <= 0) {
			report(LOG_ERR, "unrecognized keyword: \"%s\"\n", *keyvec);
		} else {
			keyvec++;
			switch (keywd->symbolcode) {
#ifdef DEBUG
			case KEYW_DUMP:
				if (*keyvec) {
					strncpy(bootpd_dump, *keyvec, sizeof(bootpd_dump));
					bootpd_dump[sizeof(bootpd_dump) - 1] = '\0';
				}
				break;
#endif
			case KEYW_GATEWAY:
				kw_gateway(keyvec);
				break;
			case KEYW_NETWORK:
				kw_network(keyvec);
				break;
			case KEYW_GCINTERVAL:
				if (*keyvec)
					gcinterval = atoi(*keyvec);
				break;
			case KEYW_FLUSHITEMS:
				if (*keyvec)
					max_cachequeue = atoi(*keyvec);
				break;
			case KEYW_FLUSHINTERVAL:
				if (*keyvec)
					flush_interval = atoi(*keyvec);
				break;
			case KEYW_BINDINGS:
				if (*keyvec)
					bindingsdir = savestr(*keyvec);
				break;
			default:
				report(LOG_ERR, "internal error: unknown keyword code\n");
				break;
			}
		}
	}
	return 0;
}


int 
check_net(gwaddr, addr)
	struct in_addr  gwaddr, addr;
{
	struct gateway *gw;
	unsigned        hashcode;

	if (gwaddr.s_addr == 0)
		gwaddr.s_addr = my_ip_addr.s_addr;
	hashcode = hash_HashFunction((u_char*)&gwaddr, 4);
	gw = (struct gateway *) hash_Lookup(gwhshtbl, hashcode, gwlcmp, &gwaddr);
	if (!gw) {
		if (debug > 2)
			report(LOG_INFO, "gateway not found %s in %s", inet_ntoa(gwaddr), conffile);
		return TRUE;
	}
	return ((gw->mask & gwaddr.s_addr) == (gw->mask & addr.s_addr));
}

/*
 * This function sanity-checks a netmask for validity.  Returns TRUE for a
 * good mask, FALSE for a bad mask.
 */
PRIVATE int 
chk_mask(mask)
	long            mask;
{
	int             count;

	mask = ntohl(mask);

	if ((mask == 0L) || (mask == -1L)) {
		return FALSE;
	}
	for (count = 32; count > 0; count--) {
		if ((mask & 0x80000000L) == 0L) {
			break;
		}
		mask <<= 1;
	}
	if (mask != 0L) {
		return FALSE;
	}
	return TRUE;
}




#define	NAME	tokenvec[0]
#define	NETWORK	tokenvec[1]
#define	MASK	tokenvec[2]
PRIVATE void 
kw_network(tokenvec)
	char          **tokenvec;
{
	struct in_addr  tmpaddr;
	struct cable   *c;
	struct gateway *gw;
	unsigned        hashcode;
	char           *cp, *last, *next, **nets, **router;
	int             i, rc, size;
	unsigned long   mask;
	struct host    *hp;

	if (!*tokenvec) {
		return;		/* Null network definition */
	}
	if (!NETWORK) {
		report(LOG_ERR, "network not specified for %s", NAME);
		return;
	}
	if (!MASK) {
		report(LOG_ERR, "mask not specified for %s", NAME);
		return;
	}
	hashcode = hash_HashFunction((u_char*)NAME, strlen(NAME));
	if (hash_Exists(cablehshtbl, hashcode, cablelcmp, NAME)) {
		report(LOG_ERR, "Network already exists, ignoring %s", NAME);
		return;
	}
	c = (struct cable *) smalloc(sizeof(struct cable));

	bzero(c, sizeof(struct cable));
	c->name = (struct shared_string *) malloc(sizeof(struct shared_string) +
						  strlen(NAME));
	c->name->linkcount = 1;
	strcpy(c->name->string, NAME);

	cp = NETWORK;
	rc = prs_inetaddr(&cp, &c->net);

	cp = MASK;
	rc = prs_inetaddr(&cp, &c->mask);


	c->code = DHCP_ACCESS;
	if (tokenvec[3]) {
		next = tokenvec[3];
		do {
			last = next;
			next = strchr(last, '|');
			if (next)
				*next++ = 0;
			if (strcmp(last, "dhcp") == 0) {
				c->code |= DHCP_ACCESS;
				c->code &= (~BOOTP_ACCESS);
			} else if (strcmp(last, "bootp") == 0) {
				c->code |= BOOTP_ACCESS;
				c->code &= (~DHCP_ACCESS);
			} else if (strcmp(last, "both") == 0)
				c->code |= BOOTP_ACCESS | DHCP_ACCESS;
			else if (strcmp(last, "roaming") == 0)
				c->code |= ROAMING_ACCESS;
			else if (strcmp(last, "registered") == 0)
				c->code |= REG_ACCESS;
			else if (strcmp(last, "any") == 0)
				c->code |= ANY;
			else
				report(LOG_ERR, "Bad keyword for arguement 4, was expecting\
			\"dhcp\",\"bootp\", or \"any\" but received \"%s\"", tokenvec[4]);
		} while (next);
	}
	c->linkcount = 1;
	hash_Insert(cablehshtbl, hashcode, nullcmp, c, c);
	mask = tmpaddr.s_addr = c->net & c->mask;

	i = 0;
	while ((tmpaddr.s_addr & c->mask) == mask) {
		hashcode = hash_HashFunction((u_char*)&tmpaddr, 4);
		hp = (struct host *) hash_Lookup(iphashtable, hashcode, ipcmp, &tmpaddr);
		if (hp && hp->flags.dynamic) {
			i++;
			hp->flags.cable = TRUE;
			hp->cable = c->name;
			c->name->linkcount++;
			if (init || (ipboundhshtbl && !hash_Exists(ipboundhshtbl, hashcode, ipcmp, &tmpaddr)))
				push_ip(c, tmpaddr);
		}
		tmpaddr.s_addr++;

	}
	if (debug > 1)
		report(LOG_INFO, "Network %s %d entries", c->name->string, i);

	router = NULL;
	if (*tokenvec[3])
		router = tokenvec + 4;
	while (router && router[0] && router[1]) {
		prs_inetaddr(router, &tmpaddr);
		hashcode = hash_HashFunction((u_char*)&tmpaddr, 4);
		gw = (struct gateway *) hash_Lookup(gwhshtbl, hashcode, gwlcmp, &tmpaddr);
		if (!gw) {
			gw = (struct gateway *) smalloc(sizeof(struct gateway));
			gw->addr.s_addr = tmpaddr.s_addr;
			size = sizeof(char *);
			gw->nets = (char **) malloc(size);
			gw->num = 1;
			gw->nets[0] = savestr(tokenvec[0]);
			gw->linkcount = 1;
			hash_Insert(gwhshtbl, hashcode, nullcmp, gw, gw);
		} else {
			size = sizeof(char *) * (gw->num + 1);
			nets = (char **) malloc(size);
			bcopy(gw->nets, nets, size - sizeof(char *));
			nets[gw->num] = savestr(tokenvec[0]);
			free(gw->nets);
			gw->num++;
		}
		prs_inetaddr(router + 1, &gw->mask);
		router += 2;
	}


	return;
}

PRIVATE void 
kw_gateway(tokenvec)
	char          **tokenvec;
{
	struct in_addr  tmpaddr;
	struct cable   *c;
	struct gateway *gw;
	unsigned        hashcode;
	char           *cp, **network;
	int             rc, size;
	unsigned long   mask;
	char          **nets;

	if (!*tokenvec) {
		return;		/* Null network definition */
	}
	if (!tokenvec[1]) {
		report(LOG_ERR, "Need Mask for gateway definition");
		return;
	}
	cp = tokenvec[0];
	prs_inetaddr(&cp, &tmpaddr);
	hashcode = hash_HashFunction((u_char*)&tmpaddr, 4);
	gw = (struct gateway *) hash_Lookup(gwhshtbl, hashcode, gwlcmp, &tmpaddr);
	if (!gw) {
		gw = (struct gateway *) smalloc(sizeof(struct gateway));
		gw->addr.s_addr = tmpaddr.s_addr;
		cp = tokenvec[1];
		prs_inetaddr(&cp, &tmpaddr);
		gw->mask = tmpaddr.s_addr;
		gw->num = 0;
		gw->linkcount = 1;
		hash_Insert(gwhshtbl, hashcode, nullcmp, gw, gw);
	}
	cp = tokenvec[1];
	prs_inetaddr(&cp, &gw->mask);
	network = tokenvec + 2;
	while (*network) {
		size = sizeof(char *) * (gw->num + 1);
		nets = (char **) malloc(size);
		bcopy(gw->nets, nets, size - sizeof(char *));
		nets[gw->num] = savestr(*network);
		gw->num++;
		free(gw->nets);
		gw->nets = nets;
		network++;
	}

	return;
}

int 
access_network(res, client)
	unsigned char   res, client;
{

	if ((BOOTP_ACCESS & client) && !(BOOTP_ACCESS & res))
		return FALSE;

	if ((ROAMING_ACCESS & res) && !(ROAMING_ACCESS & client))
		return FALSE;

	if ((REG_ACCESS & res) && !(REG_ACCESS & client))
		return FALSE;

	return TRUE;
}

struct host    *
get_dynamic_ip(gwaddr, code)
	struct in_addr  gwaddr;
	char            code;
{
	struct cable   *c;
	struct gateway *gw;
	int             rc, found, i, j;
	unsigned        hashcode;
	struct link_list *t, **l;
	struct host    *hp, *old;

	if (gwaddr.s_addr == 0)
		gwaddr.s_addr = my_ip_addr.s_addr;
	hashcode = hash_HashFunction((u_char*)&gwaddr, 4);
	gw = (struct gateway *) hash_Lookup(gwhshtbl, hashcode, gwlcmp, &gwaddr);
	if (!gw)
		return NULL;

	hp = NULL;
	found = 0;
	for (i = 0; !found; i++) {
		if (i == gw->num) {
			return NULL;
		}
		hashcode = hash_HashFunction((u_char*)gw->nets[i], strlen(gw->nets[i]));
		c = (struct cable *) hash_Lookup(cablehshtbl, hashcode, cablelcmp,
						 gw->nets[i]);
		if (!c || !c->free || !access_network(c->code, code))
			continue;

		l = &c->head;
		while (*l) {
			hashcode = hash_HashFunction((u_char*)(*l)->data, 4);
			hp = (struct host *) hash_Lookup(iphashtable, hashcode, ipcmp, (*l)->data);
			if (hash_Exists(ipboundhshtbl, hashcode, ipcmp, (*l)->data))
				report(LOG_ERR, "Tried to assign an ip which already had a binding %s", inet_ntoa(*((struct in_addr *) (*l)->data)));
			else if (hp && hp->flags.iaddr && !ping(hp->iaddr, 3, 0, 500000)) {
				report(LOG_ERR, "Address already in use %s", inet_ntoa(hp->iaddr));
			} else if (hp && hp->flags.dynamic && ((gw->mask & gw->addr.s_addr) == (gw->mask & hp->iaddr.s_addr))) {
				c->free--;
				t = *l;
				*l = (*l)->next;
				if(t == c->tail){
					c->tail=c->head;
					while(c->tail && c->tail->next)
						c->tail=c->tail->next;
				}
				free(t->data);
				free(t);
				found = 1;
				break;
			}
			l = &(*l)->next;
		}
	}

	if (hp) {
		old = (struct host *) hash_Lookup(ipboundhshtbl, hp->iphashcode, ipcmp, &hp->iaddr);
		reuse_addr(old);
	}
	return hp;
}

struct host    *
use_addr(hp, gwaddr, flag, force)
	struct host    *hp;
	struct in_addr  gwaddr;
	char            flag;
	int             force;
{
	struct cable   *c;
	struct gateway *gw;
	unsigned        hashcode;
	struct link_list *t, **l;
	int             i;
	if (!hp || !hp->flags.cable || !hp->flags.iaddr)
		return NULL;

	if (gwaddr.s_addr == 0)
		gwaddr.s_addr = my_ip_addr.s_addr;
	if (!force) {
		if (!check_net(gwaddr, hp->iaddr))
			return NULL;
		hashcode = hash_HashFunction((u_char*)&gwaddr, 4);
		gw = (struct gateway *) hash_Lookup(gwhshtbl, hashcode, gwlcmp, &gwaddr);
		if (!gw)
			return NULL;
		for (i = 0; i < gw->num; i++)
			if (strcmp(gw->nets[i], hp->cable->string) == 0)
				break;
		if (i == gw->num)
			return NULL;
	}
	hashcode = hash_HashFunction((u_char*)hp->cable->string, strlen(hp->cable->string));
	c = (struct cable *) hash_Lookup(cablehshtbl, hashcode, cablelcmp, hp->cable->string);
	if (!c)
		return hp;
	if (!access_network(c->code, flag))
		return NULL;
	l = &c->head;
	while (*l) {
		if (!bcmp((*l)->data, &hp->iaddr, 4)) {
			c->free--;
			t = *l;
			*l = (*l)->next;
			if(t == c->tail){
				c->tail=c->head;
				while(c->tail && c->tail->next)
					c->tail=c->tail->next;
			}
			free(t->data);
			free(t);
			return hp;
		}
		l = &(*l)->next;
	}

	return NULL;
}


void 
reuse_addr(hp)
	struct host    *hp;
{
	struct cable   *c;
	unsigned        hashcode;

	if (!hp || !hash_Exists(ipboundhshtbl, hp->iphashcode, ipcmp, &hp->iaddr))
		return;

	if (hp->flags.cable) {
		hashcode = hash_HashFunction((u_char*)hp->cable->string, strlen(hp->cable->string));
		c = (struct cable *) hash_Lookup(cablehshtbl, hashcode, cablelcmp, hp->cable->string);
		push_ip(c, hp->iaddr);
		ips_reclaimed++;
	}
	release_lastbind(hp);
	hashcode = hash_HashFunction((u_char*)hp->cl->data, hp->cl->length);
	leases--;
	hash_Delete(boundhshtbl, hashcode, boundcliidcmp, hp->cl, free_host);
	hash_Delete(ipboundhshtbl, hp->iphashcode, ipcmp, &hp->iaddr, free_host);
}


void 
push_ip(cable, addr)
	struct cable   *cable;
	struct in_addr  addr;
{

	if (!cable)
		return;
	if ((cable->mask & cable->net) != (cable->mask & addr.s_addr))
		return;


	if (!cable->head) {
		cable->tail = (struct link_list *) malloc(sizeof(struct link_list));
		cable->head = cable->tail;
	} else {
		cable->tail->next = (struct link_list *) malloc(sizeof(struct link_list));
		cable->tail = cable->tail->next;
	}
	bzero(cable->tail, sizeof(struct link_list));

	cable->tail->data = (char *) malloc(sizeof(struct in_addr));
	bcopy(&addr, cable->tail->data, sizeof(struct in_addr));
	cable->free++;
}



boolean 
cablelcmp(key, data)
	char           *key;
	struct cable   *data;
{
	return !strcmp(key, data->name->string);
}
boolean 
gwlcmp(key, data)
	struct in_addr *key;
	struct gateway *data;
{

	return (data->addr.s_addr == key->s_addr);
}




/*
 * This function takes a string of whitespace-separated words (tokens) and
 * returns a vector of pointers to the various tokens.  The strings pointed
 * to by the vector have leading and trailing whitespace removed and are
 * null-terminated.
 * 
 * THIS FUNCTION MODIFIES THE ARGUMENT STRING!  Nulls are inserted into the
 * original string to terminate the tokens.  This may cause orphaning of
 * memory if the caller is not careful.
 */

PRIVATE char  **
get_tokens(buff)
	char           *buff;
{
	static char    *token_vector[MAXTOKENS];
	char          **vector;
	int             i;

	i = 0;
	vector = token_vector;
	while (*buff) {
		eat_whitespace(&buff);
		if (*buff) {
			if (i < MAXTOKENS) {
				*vector++ = buff;
				i++;
				while (*buff && !isspace(*buff)) {
					buff++;
				}
				if (*buff == '\0') {
					break;
				}
				*buff++ = '\0';
			} else {
				break;
			}
		}
	}
	*vector = NULL;
	return token_vector;
}



char           *
savestr(s)
	char           *s;
{
	char           *t;

	t = malloc(strlen(s) + 1);
	if (t == NULL) {
		report(LOG_ERR, "savestr: malloc() failure -- exiting\n");
		exit(1);
	}
	strcpy(t, s);
	return t;
}
