/* 
 * Copyright 1995 Liverpool University Computer Science Department
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice appear in all copies and that both that 
 * copyright notice and this permission notice appear in supporting 
 * documentation, and that the name of L.U.C.S. not be used in advertising
 * or publicity pertaining to distribution of the software without specific, 
 * written prior permission. L.U.C.S. makes no representations about the 
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * FILE NAME:   wwwtar.c
 * CREATED:     22 Apr 1995
 * AUTHOR:      Rik Turnbull (rik@csc.liv.ac.uk)
 * DESCRIPTION: CGI program to generate HTML documents from tar files.
 *
 */

/******************************************************************************/
/******************************* INCLUDES *************************************/
/******************************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#include "common.h"

#include "wwwtar_int.h"
#include "wwwtar_config.h"

/******************************************************************************/
/****************************** DEFINITIONS ***********************************/
/******************************************************************************/
#define	WWWTAR_MIME_DEFAULT	"application/octet-stream"
#define	NUM_MIME_TYPES	(sizeof(mime_types)/sizeof(mime_types[0]))

/******************************************************************************/
/********************** INTERNAL STRUCTURES & VARIABLES ***********************/
/******************************************************************************/
typedef struct _mime_type {
    char *ending;
    char *content_type;
} mime_type;

static mime_type mime_types[] = {
    { "bin", "application/octet-stream" },
    { "oda", "application/oda" },
    { "pdf"  "application/pdf" },
    { "ai",  "application/postscript" },
    { "ps",  "application/postscript" },
    { "eps", "application/postscript" },
    { "rtf", "application/rtf" },
    { "mif", "application/x-mif" },
    { "csh", "application/x-csh" },
    { "dvi", "application/x-dvi" },
    { "hdf", "application/x-hdf" },
    { "latex", "application/x-latex" },
    { "nc", "application/x-netcdf" },
    { "cdf", "application/x-netcdf" },
    { "sh", "application/x-sh" },
    { "tcl", "application/x-tcl" },
    { "tex", "application/x-tex" },
    { "texi", "application/x-texinfo" },
    { "texinfo", "application/x-texinfo" },
    { "t", "application/x-troff" },
    { "tr", "application/x-troff" },
    { "roff", "application/x-troff" },
    { "man", "application/x-troff-man" },
    { "me", "application/x-troff-me" },
    { "src", "application/x-wais-source" },
    { "zip", "application/zip" },
    { "bcpio", "application/x-bcpio" },
    { "cpio", "application/x-cpio" },
    { "shar", "application/x-shar" },
    { "sv4cpio", "application/x-sv4cpio" },
    { "sv4crc", "application/x-sv4crc" },
    { "tar", "application/x-tar" },
    { "ustar", "application/x-ustar" },
    { "au", "audio/basic" },
    { "snd", "audio/basic" },
    { "aif", "audio/x-aiff" },
    { "aiff", "audio/x-aiff" },
    { "aifc", "audio/x-aiff" },
    { "aifc", "audio/x-aiff" },
    { "gif", "image/gif" },
    { "ief", "image/ief" },
    { "jpeg", "image/jpeg" },
    { "jpg", "image/jpeg" },
    { "jpe", "image/jpeg" },
    { "tiff", "image/tiff" },
    { "tif", "image/tiff" },
    { "ras", "image/x-cmu-raster" },
    { "pnm", "image/x-portable-anymap" },
    { "pbm", "image/x-portable-bitmap" },
    { "ppm", "image/x-portable-pixmap" },
    { "rgb", "image/x-rgb" },
    { "xbm", "image/x-xbitmap" },
    { "xpm", "image/x-xpixmap" },
    { "xwd", "image/x-xwindowdump" },
    { "html", "text/html" },
    { "htm", "text/html" },
    { "txt", "text/plain" },
    { "text", "text/plain" },
    { "rtx", "text/richtext" },
    { "tsv", "text/tab-separated-values" },
    { "etx", "text/x-setext" },
    { "mpeg", "video/mpeg" },
    { "mpg", "video/mpeg" },
    { "mpe", "video/mpeg" },
    { "qt", "video/quicktime" },
    { "mov", "video/quicktime" },
    { "avi", "video/x-msvideo" },
    { "movie", "video/x-sgi-movie" },
};

static char errmsg[256];
static char ident[]="@(#)wwwtar V1.0 (C) Liverpool University Computer Science";

/******************************************************************************/
/************************ INTERNAL FUNCTION PROTOTYPES ************************/
/******************************************************************************/
static int wwwtar_list(char *, char *, char *, int);
static int wwwtar_convert(TAR_FILE *,FILE_HEADER *, int);
static char *wwwtar_mode(mode_t, int);
static int wwwtar_null(FILE_HEADER *);
static int wwwtar_checksum(char *);
static int wwwtar_type(int, char *);
static void wwwtar_line(TAR_FILE *, char *);
static char *wwwtar_mime(char *);
static void wwwtar_header();
static void wwwtar_footer();

main(int argc, char *argv[])
{
    char *filename, *archive, *alias;
    int html = 0;

    if(argc > 1) {
        /* The 'alias' is the name of the file before its path is prepended
           and 'archive' is the full filename after.
        */
        alias = strdup(argv[1]);
        archive = strdup(fname_resolve("%s%s", WWWTAR_ROOT, alias));

        /* This can occur if the user tries a filename such as
           '../../etc/passwd' in which case we tell them to get lost.
        */
        if(strncmp(archive, WWWTAR_ROOT, strlen(WWWTAR_ROOT)) != 0) {
            wwwtar_header();
            printf("\nNice try!\n");
            wwwtar_footer();
            exit(1);
        }
    }

    /* If there are two arguments, then the usage must be
         wwwtar <archive.tar>
    */
    if(argc == 2) {
        /* We are sending HTML */
        html = 1;

        /* Send the standard header. */
        wwwtar_header();

        /* Don't extract a file. */
        filename = NULL;

        /* Print the name of the archive we are listing. */
        printf("<B>Tar Archive: %s</B>\n\n", alias);
        fflush(stdout);

        /* Generate the HTML document. */
        if(wwwtar_list(archive, alias, filename, 0) <= 0) {
            /* Looks like something went wrong. */
            printf("%s", errmsg);
            fflush(stdout);
        }
        
    /* If there are three arguments, then the usage must be
         wwwtar <archive.tar> <filename>
    */
    } else if(argc == 3) {
        filename = argv[2];
        if(wwwtar_list(archive, alias, filename, 1) == -1) {
            wwwtar_header();
            printf("%s", errmsg);
            fflush(stdout);
            html = 1;
        }
    /* If there are four arguments, then the usage must be
         wwwtar <archive.tar> <filename> text
    */
    } else if(argc == 4) {
        wwwtar_header();
        filename = argv[2];
        printf("<B>Tar Archive: %s\nFilename:    %s</B>\n", alias,
          filename);
        if(strcmp("text", argv[3]) == 0) {
            printf("</PRE><PLAINTEXT>\n");
            fflush(stdout);
            if(wwwtar_list(archive, alias, filename, 0) == -1) {
                printf("%s", errmsg);
                fflush(stdout);
            }
        } else {
            printf("\nError calling WWWTar!\n");
            html = 1;
        }

    /* More than four arguments is an error. */
    } else {
        wwwtar_header();
        printf("\nError calling WWWTar!\n");
        html = 1;
    }

    if(html)
        wwwtar_footer();
}

/* wwwtar_list:****************************************************************/
/* wwwtar_list: Lists the contents of the tar archive or extracts a file and  */
/* wwwtar_list: prints it to stdout.                                          */
/* wwwtar_list:****************************************************************/
static int wwwtar_list(char *archive, char *alias, char *filename, int header)
{
    int fp = -1, found = filename?0:1;
    FILE_HEADER file_header;
    TAR_FILE tf;
    int file_size, size, seek_size, pos = 0, read_size = 0, to_read;
    char buffer[512], *type;

    /* Make sure the tar archive isn't too large */
    file_size = stat_size(archive);
    if(file_size > WWWTAR_MAX_FILE_SIZE) {
        sprintf(errmsg, "\nTar archive is %d bytes!\nProcessing of files is restricted to %d bytes!\n", file_size, WWWTAR_MAX_FILE_SIZE);
        return(-1);
    }

    /* Open the file */
    if((fp = pipe_open(archive)) == -1) {
        sprintf(errmsg, "\nError opening %s!\n", alias);
        return(-1);
    }

    /* Read a file header */
    read_size = pipe_read(fp, (char *)&file_header, FILE_HEADER_SIZE);
    if(read_size !=  FILE_HEADER_SIZE) {
        if(read_size == 0)
           sprintf(errmsg, "\nError reading tar archive!\nEmpty file!\n");
        else
           sprintf(errmsg,"\nError reading tar archive!\n %s\n",strerror(errno));
        pipe_close(fp);
        return(-1);
    }
    pos += FILE_HEADER_SIZE;

    /* Parse headers until reach NULL record */
    while(!wwwtar_null(&file_header)) {
        int magic_ok = 0;

        /* Check the magic number - is it a tar file? */
        if(!strncmp(file_header.magic, "ustar", 5))
            magic_ok = 1;

        /* Calculate the checksum - is the archive corrupt? */
        if(!wwwtar_checksum((char *)&file_header)) {
            sprintf(errmsg, "\nError reading archive!\n");
            pipe_close(fp);
            return(-1); 
        }

        /* Convert the header */
        wwwtar_convert(&tf, &file_header, pos);

        if(filename) {
            if(strcmp(filename, tf.name) == 0) {
                found = 1;
                if(write(1, buffer, strlen(buffer)) != strlen(buffer)) {
                    sprintf(errmsg, "\nError sending data!\n");
                    pipe_close(fp);
                    return(0);
                }
                if(tf.size > 0) {
                    if(header) {
                        type = wwwtar_mime(tf.name);
                        if(type)
                            sprintf(buffer, "Content-type: %s\n\n", type);
                        else
                            sprintf(buffer, "Content-type: %s\n\n",
                              WWWTAR_MIME_DEFAULT);
                        if(write(1, buffer, strlen(buffer)) != strlen(buffer)) {
                            sprintf(errmsg, "\nError sending data!\n");
                            pipe_close(fp);
                            return(0);
                        }
                    }
                    to_read = tf.size; read_size = 0;
                    while(to_read -= read_size) {
                        /* Get size of block to read and write out */
                        if((read_size = (to_read%512)) == 0)
                            read_size = 512;
                        /* Read in a data block */
                        if(pipe_read(fp, buffer, read_size) != read_size) {
                            sprintf(errmsg, "\nError reading archive!\n");
                            pipe_close(fp);
                            return(0);
                        }
                        /* Write out block of data */
                        if(write(1, buffer, read_size) != read_size) {
                            sprintf(errmsg, "\nError sending data!\n");
                            pipe_close(fp);
                            return(0);
                        }
                    }
                    pipe_close(fp);
                    return(1);
                } else {
                    sprintf(errmsg, "\nFile has zero length!\n");
                    pipe_close(fp);
                    return(-1);
                }
            }
        }
 
        /* Skip the actual file (if there is one) */
        if(tf.size > 0) {
            seek_size = (((tf.size-1)/512)+1)*512;
            if((tf.is_ascii = wwwtar_seek(fp, tf.size, seek_size)) < 0) {
                sprintf(errmsg, "\nError reading archive!\n");
                pipe_close(fp);
                return(0); 
            }
            pos += seek_size;
        }

        if(filename == NULL) {
            /* Display the contents */
            wwwtar_line(&tf, alias);
        }

        /* Read file header for next iteration  */
        read_size = pipe_read(fp, (char *)&file_header, FILE_HEADER_SIZE);
        if(read_size != FILE_HEADER_SIZE) {
              if(read_size == 0)
                  sprintf(errmsg, "\nError reading tar archive!\nUnexpected EOF\n");
              else
                  sprintf(errmsg, "\nError reading tar archive!\n");
              pipe_close(fp);
              return(-1);
          }
          pos += FILE_HEADER_SIZE;
    }

    /* Close file */
    pipe_close(fp);

    /* Did we display a file? */
    if(!found) {
        sprintf(errmsg, "\nFile not found in tar archive!\n");
        return(-1);
    }

    /* OK */
    return(1);
}

/* wwwtar_convert:*************************************************************/
/* wwwtar_convert: This takes an actual TAR_HEADER and converts to be of type */
/* wwwtar_convert: FILE_HEADER. This is a form which is much more easier to   */
/* wwwtar_convert: handle than the standard header.                           */
/* wwwtar_convert:*************************************************************/

static int wwwtar_convert(TAR_FILE *tf,FILE_HEADER *file_header, int offset)
{
    time_t mtime;
    char *mtime_str;

    /* Copy in simple values */
    tf->mode = (mode_t)num_decimal(file_header->mode, MODE_SZ);
    tf->file_type = wwwtar_type(file_header->typeflag, file_header->name);
    tf->uid  = num_decimal(file_header->uid, UID_SZ);
    tf->gid  = num_decimal(file_header->gid, GID_SZ);
    tf->size = num_decimal(file_header->size, SIZE_SZ);
    strcpy(tf->name, file_header->name);
    strcpy(tf->linkname, file_header->linkname);
    tf->file_offset = offset;
    tf->next = NULL;

    /* Get the modification time */
    mtime = num_decimal(file_header->mtime, MTIME_SZ);
    tf->time = mtime;
    mtime_str = ctime(&mtime);
    mtime_str[16] = 0; mtime_str[24] = 0;  /* Blitz secs & newline */
    sprintf(tf->mtime, "%s %s", &mtime_str[4], &mtime_str[20]);

    /* Put file mode into drwxr--r-- form */
    strcpy(tf->permissions,wwwtar_mode(tf->mode,tf->file_type));

    /* OK */
    return(1);
}

/* wwwtar_mode:****************************************************************/
/* wwwtar_mode: Given the mode in mode_t format and the file type from the    */
/* wwwtar_mode: header this routine returns the "drwxr--r--" permission bits. */
/* wwwtar_mode:****************************************************************/

static char *wwwtar_mode(mode_t mode, int file_type)
{
    static char permissions[11];

    /* Initiliaze mem */
    strcpy(permissions, "?---------");

    switch(file_type) {
        case TYPE_NORMAL    :
        case TYPE_HARDLINK  : permissions[0] = '-'; break;
        case TYPE_SOFTLINK  : permissions[0] = 'l'; break;
        case TYPE_CHARACTER : permissions[0] = 'c'; break;
        case TYPE_BLOCK     : permissions[0] = 'b'; break;
        case TYPE_DIRECTORY : permissions[0] = 'd'; break;
        default             : permissions[0] = '?'; break;
    }

    if(mode & S_IRUSR) permissions[1] = 'r';
    if(mode & S_IWUSR) permissions[2] = 'w';
    if(mode & S_IXUSR) permissions[3] = 'x';
    if(mode & S_IRGRP) permissions[4] = 'r';
    if(mode & S_IWGRP) permissions[5] = 'w';
    if(mode & S_IXGRP) permissions[6] = 'x';
    if(mode & S_IROTH) permissions[7] = 'r';
    if(mode & S_IWOTH) permissions[8] = 'w';
    if(mode & S_IXOTH) permissions[9] = 'x';


    /* Return string */
    return(permissions);
}

/* wwwtar_null:****************************************************************/
/* wwwtar_null: A normal tar archive has 512bytes full of NULL bytes to mark  */
/* wwwtar_null: the end of the archive. This function just checks a file      */
/* wwwtar_null: header to see if it is that block.                            */
/* wwwtar_null:****************************************************************/

static int wwwtar_null(FILE_HEADER *file_header)
{
    char *bytes = (char *) file_header;
    int i = FILE_HEADER_SIZE;

    /* If we come across one byte that isn'y null then return immediately. */
    while(i--) {
        if(bytes[i] != 0)
            return(0);
    }

    /* All nulls! */
    return(1);
}

/* wwwtar_checksum:************************************************************/
/* wwwtar_checksum: This checksum is calculated by adding together all the    */
/* wwwtar_checksum: bytes in the header, with the checksum field full of      */
/* wwwtar_checksum: blanks.                                                   */
/* wwwtar_checksum:************************************************************/

static int wwwtar_checksum(char *file_header)
{
    char check_header[FILE_HEADER_SIZE];
    int val = 0, i = FILE_HEADER_SIZE;

    /* Make a copy of the header because we are going to mess it up :-) */
    memcpy(check_header, file_header, FILE_HEADER_SIZE);

    /* Mess it up by adding blanks in the checksum field */
    strncpy(&check_header[148], "        ", 8);

    /* Loop round a calculate the checksum */
    while(i--) val += check_header[i];

    /* The TEST! */
    if(val != num_decimal(((FILE_HEADER *)file_header)->chksum, CHKSUM_SZ))
        return(0);
    else
        return(1);
}

/* wwwtar_type:****************************************************************/
/* wwwtar_type: If we get a UNIX file type in the header, then we must decide */
/* wwwtar_type: if it is a directory or a normal file by looking at its file  */
/* wwwtar_type: name.                                                         */
/* wwwtar_type:****************************************************************/

static int wwwtar_type(int file_type, char *name)
{
    /* Test file type - change it, if set to TYPE_UNIX */
    if(file_type == TYPE_UNIX) {
        if(name[strlen(name) - 1] == '/')
            return(TYPE_DIRECTORY);
        else
            return(TYPE_NORMAL);
    }
    else {
        return(file_type);
    }
}

/* wwwtar_line:****************************************************************/
/* wwwtar_line: Display an entry in the tar file.
/* wwwtar_line:****************************************************************/

static void wwwtar_line(TAR_FILE *tf, char *alias)
{
    char buffer[BUFSIZ];
    static int title = 0;

    if(title == 0) {
        printf("%9s %10s %-17s %s<HR>","Permissions","Size","Date", "Name");
        title = 1;
    }
    switch(tf->file_type) {
        case TYPE_SOFTLINK:
            sprintf(buffer, "%s  %10d %s %s -> %s", tf->permissions, tf->size, tf->mtime, tf->name, tf->linkname);
            break;
        case TYPE_HARDLINK:
            sprintf(buffer, "%s  %10d %s %s linked to %s", tf->permissions, tf->size, tf->mtime, tf->name, tf->linkname);
              break;
        case TYPE_DIRECTORY:
            sprintf(buffer, "%s  %10d %s %s", tf->permissions, tf->size, tf->mtime, tf->name);
            break;
        default:
            if(tf->is_ascii && wwwtar_mime(tf->name) == NULL)
                sprintf(buffer, "%s  %10d %s <A HREF=/%s/wwwtar?%s+%s+text>%s</A>", tf->permissions, tf->size, tf->mtime, WWWTAR_CGIBIN, alias, tf->name, tf->name);
            else
                sprintf(buffer, "%s  %10d %s <A HREF=/%s/wwwtar?%s+%s BROWSER=ignore>%s</A>", tf->permissions, tf->size, tf->mtime, WWWTAR_CGIBIN, alias, tf->name, tf->name);
    }

    printf("%s\n", buffer);
}

/* wwwtar_line:****************************************************************/
/* wwwtar_line: Duplication of pipe_seek which checks to see if the block is  */
/* wwwtar_line: ascii text.                                                   */
/* wwwtar_line:****************************************************************/

int wwwtar_seek(int fp, int size, int offset)
{
    char *ptr, buffer[BUFSIZ];
    int total = 0, is_ascii = 1, ascii = 0, readbytes = 0, val, i;

    /* We must actually perform a read and discard the data, rather than do
       an acutual seek - apparently.
    */
    while(offset > 0) {
        val = offset > BUFSIZ ? BUFSIZ : offset;
        while(((readbytes = read(fp, buffer, val)) > 0) && (val > 0)) {
            if(total < size) {
                i = readbytes; ptr = buffer;
                while(total < size && i--) {
                    if(isascii(*ptr++))
                        ascii++;
                    total++;
                }
            }
            val -= readbytes;
        }
        if(readbytes == -1)
            return(-1);
        offset -= BUFSIZ;
    }

    if(total > 0)
        is_ascii = ((int)((float)ascii/(float)total * 100.0))==100?1:0;

    return(is_ascii);
}

/* wwwtar_mime:****************************************************************/
/* wwwtar_mime: Search for the appropriate MIME type, for the filename.       */
/* wwwtar_mime:****************************************************************/

static char *wwwtar_mime(char *filename)
{
    int i;
    char *ptr, *type = NULL;

    ptr = strrchr(filename, '.');
    if(ptr) {
        ptr++;
        for(i = 0; i < NUM_MIME_TYPES; i++) {
            if(strcmp(ptr, mime_types[i].ending) == 0) {
                type = mime_types[i].content_type;
                break;
            }
        }
    }

    return(type);
}

/* wwwtar_header:**************************************************************/
/* wwwtar_header: Print the standard CGI script header.                       */
/* wwwtar_header:**************************************************************/

static void wwwtar_header()
{
    printf("Content-type: text/html\n\n");
    printf("<TITLE>WWWTar Query</TITLE>\n");
    printf("%s\n", WWWTAR_DOC_HEADER);
    printf("<PRE>\n");

    fflush(stdout);
}

/* wwwtar_header:**************************************************************/
/* wwwtar_header: Print a footer. Boring really.                              */
/* wwwtar_header:**************************************************************/

static void wwwtar_footer()
{
    printf("%s\n", WWWTAR_FOOTER);
    fflush(stdout);
}
